package mv3302;

import java.awt.geom.Point2D;
import static mv3302.SimplestMover.NaP;
import simkit.SimEntityBase;
import simkit.smd.Mover;

/**
 * Simple Path Mover Manager. Directs a listening Mover to each wayPoint in
 * order until the last, when it "stops"
 *
 * @author ahbuss
 */
public class SimplePathMoverManager extends SimEntityBase {

    /**
     * If true, start immediately
     */
    private boolean startOnRun;

    /**
     * Given list of wayPoints
     */
    private Point2D[] path;

    /**
     * index of next wayPoint
     */
    protected int nextIndex;

    /**
     * next destination
     */
    protected Point2D nextDestination;

    /**
     * Instantiate nextDestination as copy of NaP
     */
    public SimplePathMoverManager() {
        nextDestination = (Point2D) NaP.clone();
    }

    /**
     * Instantiate with the given parameters
     *
     * @param path Given path
     * @param startOnRun if true, start on Run event
     */
    public SimplePathMoverManager(Point2D[] path, boolean startOnRun) {
        this();
        setPath(path);
        setStartOnRun(startOnRun);
    }

    /**
     * Set nextIndex to -1<br>
     * set nextDestination to NaP coordinates
     */
    @Override
    public void reset() {
        super.reset();
        nextIndex = -1;
        nextDestination.setLocation(NaP);
    }

    /**
     * If startOnRun is true, schedule NextWayPoint
     */
    public void doRun() {
        firePropertyChange("nextIndex", getNextIndex());

        if (isStartOnRun()) {
            waitDelay("NextWayPoint", 0.0);
        }
    }

    /**
     * Increment nextIndex<br>
     * set nextDestination to path[nextIndex]<br>
     * schedule MoveTo with nextDestination
     */
    public void doNextWayPoint() {
        int oldNextIndex = getNextIndex();
        nextIndex += 1;
        firePropertyChange("nextIndex", oldNextIndex, getNextIndex());

        Point2D oldDestination = getNextDestination();
        nextDestination.setLocation(path[nextIndex]);
        firePropertyChange("nextDestination", oldDestination, getNextDestination());

        waitDelay("MoveTo", 0.0, getNextDestination());
    }

    /**
     * If more way points on path, schedule NextWayPoint<br>
     * Else schedule OrderStop
     *
     * @param mover Given Mover (should be the one this is listening to)
     */
    public void doEndMove(Mover mover) {
        if (nextIndex < path.length - 1) {
            waitDelay("NextWayPoint", 0.0);
        }

        if (nextIndex == path.length - 1) {
            waitDelay("OrderStop", 0.0);
        }
    }

    /**
     * @return the startOnRun
     */
    public boolean isStartOnRun() {
        return startOnRun;
    }

    /**
     * @param startOnRun the startOnRun to set
     */
    public void setStartOnRun(boolean startOnRun) {
        this.startOnRun = startOnRun;
    }

    /**
     * @return topy of he path
     */
    public Point2D[] getPath() {
        return path.clone();
    }

    /**
     * 
     * @param index Given index
     * @return path[index]
     */
    public Point2D getPath(int index) {
        return path[index];
    }

    /**
     * @param path the path to set
     */
    public void setPath(Point2D[] path) {
        this.path = path.clone();
    }

    /**
     * 
     * @param index Given index
     * @param path Given waypoint
     */
    public void setPath(int index, Point2D path) {
        this.path[index] = path;
    }

    /**
     * @return the nextIndex
     */
    public int getNextIndex() {
        return nextIndex;
    }

    /**
     * @return copy of the nextDestination
     */
    public Point2D getNextDestination() {
        return (Point2D) nextDestination.clone();
    }

}
